// Example Matrix Multiplication with Posix Threads
// Use "gcc filename.c -lpthread" to compile and "./a.out" to run
//
// Read Section 11.6.9, pages 418-419 from Rago and Stevens books
// on barrier synchronization

// LAB: Experiment further with barriers
// Example: Try to include load balancing phase in between row
// computations --  remember to use pthread setaffinity functions

#define NO_ITERATIONS 200	// use many runs for better timing accuracy
#define MAX_NUMBER 10		// matrix with random integers (0 to 9)
#define N        15		// matrix size
#define NPROC 5			// number of threads (there is no parent)
#define ROWS (N/NPROC)

#include <pthread.h>	// use Posix threads
#include <unistd.h> 
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <sys/times.h> // timers

pthread_t matrixThread[NPROC]; // create NPROC threads with pthread_create
void *MatrixProduct (void *);  // each pthread will run function MatrixProduct which has one argument

pthread_barrier_t bar1;
pthread_barrier_t bar2;
pthread_barrier_t bar3;

struct whatever{
		int A[N][N];
		int B[N][N];
		int C[N][N];
}*myvars;

// function that stores random integers from 0 to MAX_NUMBER-1 in an array
int readmat(int mat[N][N],int n)
{
	long i,j;

	for (i=0;i<n;i++)
		for (j=0;j<n;j++) {
			mat[i][j]= rand() % MAX_NUMBER;
	}
	return(0);
}

void printArrays(int A[N][N],int B[N][N],int C[N][N], int size) {
	int i, j;

	printf("\n");
	fflush(stdout);

	// optional printing of results
	for (i=0; i< size; i++) {

		if (i==0) {
			printf("    A ");
			fflush(stdout);
		} else {
			printf("      ");
			fflush(stdout);
		}
		for (j=0; j< size; j++) {
			printf("%d ", myvars->A[i][j]);
			fflush(stdout);
		}
		if (i==0) {
			printf("  B ");
			fflush(stdout);
		} else {
			printf("    ");
			fflush(stdout);
		}
		for (j=0; j< size; j++) {
			printf("%d ", myvars->B[i][j]);
			fflush(stdout);
		}
		if (i==0) {
			printf("  C ");
			fflush(stdout);
		} else {
			printf("    ");
			fflush(stdout);
		}
		for (j=0; j< size; j++) {
			printf("%d ", myvars->C[i][j]);
			fflush(stdout);
		}
		printf("\n");
		fflush(stdout);
	}

	printf("\n");
	fflush(stdout);
}

int main(int argc,char *argv[])
{

	int myid=0,memid; // myid and index to segment
	int i,j,k,x,y,l,no_runs;
	int sum=0;

	// timing variables
	double ticspersec;
	double t1,t2,avg_time=0.0;
	double sum_avg_time_all_iterations=0;
	int no_iterations,delay;
	struct tms tb;

	// thread attributes structure
	pthread_attr_t attr;

	srand(1L); 

	for (no_runs=0; no_runs<NO_ITERATIONS; no_runs++)
	{
		pthread_barrier_init(&bar1, NULL, NPROC+1);
		pthread_barrier_init(&bar2, NULL, NPROC);
		pthread_barrier_init(&bar3, NULL, NPROC+1);

		// allocate shared memory with malloc
		myvars = (struct whatever*)malloc (sizeof(struct whatever));
		
		if(readmat(myvars->A,N)<0)
			  exit(1+printf("File has negative arguments.\n"));

		if(readmat(myvars->B,N)<0)
			  exit(1+printf("File has negative arguments.\n"));

		pthread_attr_init(&attr);

		for(i=myid;i<NPROC;i++){
			// create NPROC processes that will run function MatrixProduct with argument i
			printf("in pthread_create myid:%d \n", i);
			pthread_create(&matrixThread[i],&attr,MatrixProduct,(void *)(intptr_t) i);
		}

		// use barrier to precisely measure timing (all threads)
		pthread_barrier_wait(&bar1);
	 
	  	// compute how many clock ticks correspond to 1 sec
		ticspersec=(double)sysconf(_SC_CLK_TCK);

		// start timing for the parallel part
		t1=(double)times(&tb);

		// use barrier to precisely measure timing (all threads)
		pthread_barrier_wait(&bar3);

		// create default attributes for thread (stack size, scheduling algorithm)
		//  end of timing parallel part (compute average)
		t2=(double)times(&tb);
		avg_time+=(t2-t1)/ticspersec;
		sum_avg_time_all_iterations += avg_time;

		// wait until all threads terminate
		for(i=myid;i<NPROC;i++)
		{
			pthread_join(matrixThread[i],NULL);
		}

	}
	printf("Average time (C=A*B): %lf sec \n", sum_avg_time_all_iterations/NO_ITERATIONS);
	fflush(stdout);

	// printing of results
	printArrays(myvars->A, myvars->B, myvars->C, N);

	pthread_barrier_destroy(&bar1);
	pthread_barrier_destroy(&bar2);
	pthread_barrier_destroy(&bar3);

	free (myvars);
	pthread_exit(NULL);

	return 0;
}
	 
// function called by all threads
void *MatrixProduct (void *arg)
{

	// local variables, different for each thread
	int  i, j, k, l, m, first,last;
	int id = (intptr_t) arg;
	int sum;
	int rc; // return code from pthread_barrier_wait

	pthread_barrier_wait(&bar1); // sync with main process for start measuring time
	
	if (i==0)
		system("sleep 0.0005");

	// calculating array C
	for (i=id;i<N;i+=NPROC) {
		//printf("%d goes in array computation loop -> ROWS:(%d %d ... %d) \n", id, id, id+NPROC, N);
		for (j=0;j<N;j++) {
			sum=0;
			for (k=0;k<N;k++) {
				// sum is a local variable, so no lock is necessary
				// A and B are read so no lock is necessary 
	 			sum += (myvars->A[i][k])*(myvars->B[k][j]); 
			}
			// Since a different thread writes each C[i][j] for any i, j
			// and only a single write occurs, there is no need for lock
			myvars->C[i][j] = sum;
		}

		rc = pthread_barrier_wait(&bar2); //sync for intermediate timing (threads only)
		if (rc == PTHREAD_BARRIER_SERIAL_THREAD) { // slowest thread
			// ... example: e.g. compute delays and use them for dynamic load balancing
			// printing results
			printArrays(myvars->A, myvars->B, myvars->C, N);
		}

		rc = pthread_barrier_wait(&bar2); //sync for intermediate timing (threads only)


		//printf("%d goes on before end loop %d \n", id);

 	}

	//printf("%d goes in end loop %d \n", id);


	pthread_barrier_wait(&bar3); // sync with main process for end measuring time

}
