// Condition variables for sending signals among processes
// (type pthread_cond_var_t, ops: pthread_cond_signal & pthread_cond_wait)
//
// a) a "good" pair (threads 1, 2) tries to keep a variable count from
// COUNT_HALT1 to COUNT_HALT2
// b) another "bad" pair (threads 3, 4) tries to push the value < COUNT_HALT1
// c) another "bad" pair (threads 5, 6) tried to push the value > COUNT_HALT2
// Every pair of threads consists of one thread checking the condition and 
// another changing count (and monitoring). Interaction is through functions
// pthread_cond_signal() < ->  pthread_cond_wait() (before changing)
// These functions must be called within mutexes, since both change same memory state of the condition variable

// Extension 1: Can you add another good pair of threads?
// Extension 2: Use pthread_cond_bcast to perform simultaneous actions, e.g. extend pairs to triplets?
// Extension 3: Can you reduce the number of lost pthread_cond_signals (not captured by pthread_cond_wait)?

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

#define COUNT_DONE  10000  // upper limit for termination, out (0,1000)
#define COUNT_HALT1 3000   // lower threshold (if below, enabling good thread)
#define COUNT_HALT2 7000   // upper threshold (if above, enabling good thread)
#define MID_POINT ((COUNT_HALT1+COUNT_HALT2)/2)  // mesos oros 

// set initial count
int  count = 5000;

// 6 threads (3 pairs: good, bad, bad)
pthread_t thread1, thread2, thread3, thread4, thread5, thread6;
void *functionCount1();
void *functionCount2();
void *functionCount3();
void *functionCount4();
void *functionCount5();
void *functionCount6();

// mutex for count change performed by processes 1, 3 or 5
pthread_mutex_t count_mutex     = PTHREAD_MUTEX_INITIALIZER;

// condition variables & mutex pairs, one for each pair of processes:
// (1,2), (3,4), (5,6)

pthread_mutex_t condition_mutex1 = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t  condition_cond1  = PTHREAD_COND_INITIALIZER;

pthread_mutex_t condition_mutex2 = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t  condition_cond2  = PTHREAD_COND_INITIALIZER;

pthread_mutex_t condition_mutex3 = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t  condition_cond3  = PTHREAD_COND_INITIALIZER;


int main() {
	pthread_create(&thread1, NULL, &functionCount1, NULL);
	pthread_create(&thread2, NULL, &functionCount2, NULL);
	pthread_create(&thread3, NULL, &functionCount3, NULL);
	pthread_create(&thread4, NULL, &functionCount4, NULL);
	pthread_create(&thread5, NULL, &functionCount5, NULL);
	pthread_create(&thread6, NULL, &functionCount6, NULL);

	pthread_join(thread1, NULL);
	pthread_join(thread2, NULL);
	pthread_join(thread3, NULL);
	pthread_join(thread4, NULL);
	pthread_join(thread5, NULL);
	pthread_join(thread6, NULL);

  return 0;

}

// good pair (threads 1, 2) aims to keep count from COUNT_HALT1 to COUNT_HALT2
void *functionCount1()
{
	printf("IN_THREAD1 \n");
	fflush(stdout);
	for(;;) {

		// exit if count is off limits
		if ((count >= COUNT_DONE)||(count < 0)) return(NULL);

      		// this thread waits for a signal from the monitoring thread 
      		// note: condition variables are always used inside locks
      		pthread_mutex_lock(&condition_mutex1);

		printf("THREAD1_wait - count:%d \n",count);
		fflush(stdout);
            		pthread_cond_wait(&condition_cond1, &condition_mutex1);

		pthread_mutex_unlock(&condition_mutex1);

		printf("THREAD1_accesses count_mutex \n");
		fflush(stdout);

      		// change count
		pthread_mutex_lock(&count_mutex);

			if (count < MID_POINT)
				count++;
			else if (count > MID_POINT)
				count--;

			printf("THREAD1_action - count: %d\n",count);
			fflush(stdout);

		pthread_mutex_unlock(&count_mutex);
	}
}

void *functionCount2()
{
	for(;;) {
		// exit if count is off limits
		if ((count >= COUNT_DONE)||(count < 0)) return(NULL);

		pthread_mutex_lock(&condition_mutex1);
			pthread_cond_signal(&condition_cond1);
			printf("THREAD2_signal - count:%d\n", count);
			fflush(stdout);
		pthread_mutex_unlock(&condition_mutex1);
    	}
}

// bad pair (threads 3, 4) aims to push count below COUNT_HALT1
void *functionCount3()
{
	printf("IN_THREAD3 \n");
	fflush(stdout);
	for(;;) {

		// exit if count is off limits
		if ((count >= COUNT_DONE)||(count < 0)) return(NULL);

      		// this thread waits for a signal from the monitoring thread 
      		// note: condition variables are always used inside locks
      		pthread_mutex_lock(&condition_mutex1);

		printf("THREAD3_wait - count:%d \n",count);
		fflush(stdout);
            		pthread_cond_wait(&condition_cond1, &condition_mutex1);

		pthread_mutex_unlock(&condition_mutex1);

		printf("THREAD3_accesses count_mutex \n");
		fflush(stdout);

      		// change count
		pthread_mutex_lock(&count_mutex);

			if (count <= MID_POINT)
				count--;

			printf("THREAD3_action - count: %d\n",count);
			fflush(stdout);

		pthread_mutex_unlock(&count_mutex);
	}
}

void *functionCount4()
{
	for(;;) {
		// exit if count is off limits
		if ((count >= COUNT_DONE)||(count < 0)) return(NULL);

		if (count <= MID_POINT) {
			pthread_mutex_lock(&condition_mutex1);
				pthread_cond_signal(&condition_cond1);
				printf("THREAD4_signal - count:%d\n", count);
				fflush(stdout);
			pthread_mutex_unlock(&condition_mutex1);
    		}
	}
}


// bad pair (threads 5, 6) aims to push count below COUNT_HALT1
void *functionCount5()
{
	printf("IN_THREAD5 \n");
	fflush(stdout);
	for(;;) {

		// exit if count is off limits
		if ((count >= COUNT_DONE)||(count < 0)) return(NULL);

      		// this thread waits for a signal from the monitoring thread 
      		// note: condition variables are always used inside locks
      		pthread_mutex_lock(&condition_mutex1);

		printf("THREAD5_wait - count:%d \n",count);
		fflush(stdout);
            		pthread_cond_wait(&condition_cond1, &condition_mutex1);

		pthread_mutex_unlock(&condition_mutex1);

		printf("THREAD5_accesses count_mutex \n");
		fflush(stdout);

      		// change count
		pthread_mutex_lock(&count_mutex);

			if (count >= MID_POINT)
				count++;

			printf("THREAD5_action - count: %d\n",count);
			fflush(stdout);

		pthread_mutex_unlock(&count_mutex);
	}
}

void *functionCount6()
{
	for(;;) {
		// exit if count is off limits
		if ((count >= COUNT_DONE)||(count < 0)) return(NULL);

		if (count >= MID_POINT) {
			pthread_mutex_lock(&condition_mutex1);
				pthread_cond_signal(&condition_cond1);
				printf("THREAD6_signal - count:%d\n", count);
				fflush(stdout);
			pthread_mutex_unlock(&condition_mutex1);
    		}
	}
}
